//==========================================================================;
//
// Written by Grzegorz Lewandowski
//
//
//--------------------------------------------------------------------------;
//
// NPAC Audio Coder
//
// Encodes an audio stream into an NPAC Audio format
//

#include <streams.h>
#include <initguid.h>
#include <uuids.h>

#include "NPACuids.h"
#include "adpcm.h"
#include "NPACACod.h"

#include "DbgLvl.h"


//
// setup data
//
AMOVIESETUP_MEDIATYPE sudInPinTypes =   { &MEDIATYPE_Audio,              // clsMajorType
                                          &MEDIASUBTYPE_NULL };           // clsMinorType

AMOVIESETUP_MEDIATYPE sudOutPinTypes =  { &MEDIATYPE_Stream,
                                          &MEDIASUBTYPE_NULL };

AMOVIESETUP_PIN psudPins[] = { { L"Input"			// strName
                               , FALSE				// bRendered
                               , FALSE				// bOutput
                               , FALSE				// bZero
                               , FALSE				// bMany
                               , &CLSID_NULL		// clsConnectsToFilter
                               , L"Output"			// strConnectsToPin
                               , 1					// nTypes
                               , &sudInPinTypes }	// lpTypes
                             , { L"Output"			// strName
                               , FALSE				// bRendered
                               , TRUE				// bOutput
                               , FALSE				// bZero
                               , FALSE				// bMany
                               , &CLSID_NULL		// clsConnectsToFilter
                               , L"Input"			// strConnectsToPin
                               , 1					// nTypes
                               , &sudOutPinTypes } };// lpTypes


AMOVIESETUP_FILTER sudNPAC = { &CLSID_NPACAudioCoder,             // clsID
                               L"NPAC Audio Coder",               // strName
                               MERIT_DO_NOT_USE,                  // dwMerit
                               2,	                              // nPins
                               psudPins };                        // lpPin



// COM Global table of objects in this dll
CFactoryTemplate g_Templates[ 1 ] = { { L"NPAC Audio Coder",
                                        &CLSID_NPACAudioCoder,
										CNPACAudioCoder::CreateInstance } };

// Count of objects listed in g_cTemplates
int g_cTemplates = 1;


//
// Constructor
//
CNPACAudioCoder::CNPACAudioCoder( TCHAR *tszName, LPUNKNOWN punk, HRESULT *phr )
			   : CTransformFilter( tszName, punk, CLSID_NPACAudioCoder, phr )
{
	DbgSetModuleLevel( LOG_ERROR | LOG_TRACE, LOWEST );
	m_uFPS = 10;
	DbgLog(( LOG_TRACE, HIGHEST, TEXT( "NPAC Audio Coder created" ) ));
}


//
// Destructor
//
CNPACAudioCoder::~CNPACAudioCoder( void )
{
	DbgLog(( LOG_TRACE, HIGHEST, TEXT( "CNPACAudioCoder destroyed" ) ));
}

//
// CreateInstance
//
CUnknown *CNPACAudioCoder :: CreateInstance( LPUNKNOWN punk, HRESULT *phr )
{
    CNPACAudioCoder *pNewObject = new CNPACAudioCoder( NAME( "NPAC Audio Coder" ), punk, phr );
    if ( pNewObject == NULL )
	{
        *phr = E_OUTOFMEMORY;
    }

    return pNewObject;
}


//
// GetSetupData
//
LPAMOVIESETUP_FILTER CNPACAudioCoder::GetSetupData()
{
  return &sudNPAC;
}


//
// Transform
//
HRESULT CNPACAudioCoder::Transform( IMediaSample *pIn, IMediaSample *pOut )
{
	DbgLog(( LOG_TRACE, LOWEST, TEXT( "CNPACAudioCoder::Transform" ) ));
	PBYTE pInput;
	PBYTE pOutput;

	HRESULT hr = pIn->GetPointer( &pInput );
	if( FAILED( hr ) )
	{
		DbgLog(( LOG_ERROR, HIGHEST, TEXT( "Transform: INPUT SAMPLE ERROR" ) ));
		return E_UNEXPECTED;
	}
	hr = pOut->GetPointer( &pOutput );
	if( FAILED( hr ) )
	{
		DbgLog(( LOG_ERROR, HIGHEST, TEXT( "Transform: OUTPUT SAMPLE ERROR" ) ));
		return E_UNEXPECTED;
	}

	unsigned uProcessed = 0;
	// Get size of input sample
	unsigned uInputSize = pIn->GetActualDataLength();
	unsigned uOutputSize = 0;
	if( m_bFirst )
	{
		AddHeader( pOutput);
		pOutput += 26;
		uOutputSize += 26;
		m_bFirst = FALSE;
	}
	DbgLog(( LOG_TRACE, LOWEST, TEXT( "Transform: INPUT BUFFER %lu bytes" ), uInputSize ));
	// Copy input sample to buffer
	memcpy( m_pBuffer + m_uBufferSize, pInput, uInputSize );
	m_uBufferSize += uInputSize;
	DbgLog(( LOG_TRACE, LOWEST, TEXT( "Transform: BUFFER SIZE %lu bytes" ), m_uBufferSize ));
	// if enough data process frame
	while( m_uBufferSize >= m_uFrameSize )
	{
		m_CoderState.valprev = 0;
		m_CoderState.index = 0;
		adpcm_coder( (short*)( m_pBuffer + uProcessed ),(char*)pOutput, m_uFrameSize / 2, &m_CoderState );
		DbgLog(( LOG_TRACE, LOWEST, TEXT( "Transform: Coding %lu bytes" ), m_uFrameSize ));
		m_uBufferSize -= m_uFrameSize;
		uProcessed += m_uFrameSize;
	}
	DbgLog(( LOG_TRACE, LOWEST, TEXT( "Transform: %lu bytes processed" ), uProcessed ));
	// reshape buffer
	if( uProcessed > 0 )
	{
		memcpy( m_pBuffer, m_pBuffer + uProcessed, m_uBufferSize );
	}
	uOutputSize += uProcessed / 4;
	pOut->SetActualDataLength( uOutputSize );
	DbgLog(( LOG_TRACE, LOWEST, TEXT( "Transform: OUTPUT BUFFER %lu bytes" ), uOutputSize ));
	return NOERROR;
}



//
// CheckInputType
//
HRESULT CNPACAudioCoder::CheckInputType( const CMediaType *mtIn )
{
    DbgLog(( LOG_TRACE, HIGH, TEXT( "CNPACAudioCoder::CheckInputType" ) ));
	DbgLog(( LOG_TRACE, HIGH, TEXT( "Major type %s" ), GuidNames[ mtIn->majortype ] ));
    DbgLog(( LOG_TRACE, MEDIUM, TEXT( "Subtype %s" ), GuidNames[ mtIn->subtype ] ));
	
	WAVEFORMATEX *pwfx = (WAVEFORMATEX*)mtIn->pbFormat;

    DbgLog(( LOG_TRACE, MEDIUM, TEXT( "Fixed size sample %d" ), mtIn->bFixedSizeSamples ));
    DbgLog(( LOG_TRACE, MEDIUM, TEXT( "Temporal compression %d" ), mtIn->bTemporalCompression ));
	DbgLog(( LOG_TRACE, MEDIUM, TEXT( "Sample size %d" ), mtIn->lSampleSize ));
    DbgLog(( LOG_TRACE, MEDIUM, TEXT( "Format size %d" ), mtIn->cbFormat ));

    if( mtIn->majortype != MEDIATYPE_Audio)
	{
        DbgLog(( LOG_TRACE, HIGH, TEXT( "CheckInputType: NOT AUDIO TYPE" ) ));
		return E_INVALIDARG;
    }

    // Now when we know it's audio we can
    // dump the contents of the WAVEFORMATEX type-specific format structure

    DbgLog(( LOG_TRACE, MEDIUM, TEXT( "wFormatTag %u" ), pwfx->wFormatTag ));
    DbgLog(( LOG_TRACE, MEDIUM, TEXT( "nChannels %lu" ), pwfx->nChannels ));
    DbgLog(( LOG_TRACE, MEDIUM, TEXT( "nSamplesPerSec %lu" ), pwfx->nSamplesPerSec ));
    DbgLog(( LOG_TRACE, MEDIUM, TEXT( "nAvgBytesPerSec %lu" ), pwfx->nAvgBytesPerSec ));
    DbgLog(( LOG_TRACE, MEDIUM, TEXT( "nBlockAlign %lu" ), pwfx->nBlockAlign ));
    DbgLog(( LOG_TRACE, MEDIUM, TEXT( "wBitsPerSample %u" ), pwfx->wBitsPerSample ));

    // PCM uses a WAVEFORMAT and does not have the extra size field

    if( mtIn->cbFormat >= sizeof(WAVEFORMATEX) )
	{
        DbgLog(( LOG_TRACE, MEDIUM, TEXT( "cbSize %u" ), pwfx->cbSize ));
    }

    // Reject invalid format blocks

    if( mtIn->cbFormat < sizeof(PCMWAVEFORMAT) )
	{
		DbgLog(( LOG_TRACE, HIGH, TEXT( "CheckInputType: INVALID FORMAT BLOCK" ) ));
        return E_INVALIDARG;
	}

    // Reject compressed audio

    if( mtIn->bTemporalCompression )
	{
		DbgLog(( LOG_TRACE, HIGH, TEXT( "CheckInputType: COMPRESSED AUDIO" ) ));
        return E_INVALIDARG;
    }

    // Accept only 16 bit

    if( pwfx->wBitsPerSample != 16 )
	{
		DbgLog(( LOG_TRACE, HIGH, TEXT( "CheckInputType: ONLY 16 BITS PER SAMPLE" ) ));
        return E_INVALIDARG;
    }

    return NOERROR;

}


//
// CheckTransform
//
HRESULT CNPACAudioCoder :: CheckTransform( const CMediaType *mtIn,
                                           const CMediaType *mtOut )
{
    if( ( mtIn->majortype == MEDIATYPE_Audio ) &&
		( mtOut->majortype == MEDIATYPE_Stream ) )
	{
        DbgLog(( LOG_TRACE, LOWEST, TEXT( "CheckTransform: OK" ) ));
		return NOERROR;
    }
	DbgLog(( LOG_TRACE, LOW, TEXT( "CheckTransform: FAILED" ) ));
    return E_FAIL;
}


//
// DecideBufferSize
//
HRESULT CNPACAudioCoder :: DecideBufferSize( IMemAllocator *pAlloc,
                                             ALLOCATOR_PROPERTIES *pProperties )
{
	DbgLog(( LOG_TRACE, LOW, TEXT( "CNPACAudioCoder::DecideBufferSize" ) ));
    if ( m_pInput == NULL )
	{
        return E_UNEXPECTED;
    }

    ASSERT(pAlloc);
    ASSERT(pProperties);
    HRESULT hr = NOERROR;

	pProperties->cBuffers = 1;
    pProperties->cbBuffer = 16 * 1024;

	ASSERT(pProperties->cbBuffer);

    
    // Ask the allocator to reserve us some sample memory, NOTE the function
    // can succeed (that is return NOERROR) but still not have allocated the
    // memory that we requested, so we must check we got whatever we wanted

    ALLOCATOR_PROPERTIES Actual;
    hr = pAlloc -> SetProperties( pProperties, &Actual );
    if ( FAILED( hr ) )
	{
        return hr;
		DbgLog(( LOG_ERROR, HIGHEST, TEXT( "DecideBufferSize: FAILED" ) ));
    }

    ASSERT( Actual.cBuffers == 1 );

    if ( pProperties -> cBuffers > Actual.cBuffers ||
         pProperties -> cbBuffer > Actual.cbBuffer )
	{
		return E_FAIL;
		DbgLog(( LOG_ERROR, HIGHEST, TEXT( "DecideBufferSize: FAILED" ) ));
    }
	DbgLog(( LOG_TRACE, LOW, TEXT( "DecideBufferSize: BUFFER %lu bytes" ), Actual.cbBuffer ));
    return NOERROR;
}


//
// GetMediaType
//
HRESULT CNPACAudioCoder::GetMediaType( int iPosition, CMediaType *pMediaType )
{
	// filter supports only one type
	if ( iPosition < 0 )
	{
		DbgLog(( LOG_ERROR, HIGH, TEXT( "GetMediaType: INVALID ARGUMENT" ) ));
		return E_INVALIDARG; // iPosition negative - no sense at all
    }
    if ( iPosition > 0 )
	{
		DbgLog(( LOG_TRACE, LOW, TEXT( "GetMediaType: NO MORE SUPPORTED TYPES" ) ));
		return VFW_S_NO_MORE_ITEMS; // no more supported types - sorry!
    }

	pMediaType->majortype = MEDIATYPE_Stream;
	pMediaType->subtype = MEDIASUBTYPE_NPAC;
	pMediaType->bFixedSizeSamples = TRUE;
	pMediaType->bTemporalCompression = TRUE;

	DbgLog(( LOG_TRACE, LOW, TEXT( "GetMediaType: SAMPLE SIZE %lu BYTES" ), pMediaType->lSampleSize ));
	
	return NOERROR;
}


//
// StartStreaming
//
HRESULT CNPACAudioCoder::StartStreaming( void )
{
	CMediaType mtInput;
	mtInput = m_pInput->CurrentMediaType();
	WAVEFORMATEX *pwfx = (WAVEFORMATEX*)mtInput.pbFormat;
	m_bFirst = TRUE;
	m_uFrameSize = pwfx->nAvgBytesPerSec / m_uFPS;
	m_uFrameSize -= m_uFrameSize % 4;
	DbgLog(( LOG_TRACE, MEDIUM, TEXT( "Frame size is %lu" ), m_uFrameSize ));
	m_pBuffer = new BYTE[ 3 * m_uFrameSize ];
	m_uBufferSize = 0;
	m_uRemainder = 0;
	return NOERROR;
}


//
// AddHeader
//
void CNPACAudioCoder::AddHeader( PBYTE pOut )
{
	DbgLog(( LOG_TRACE, LOW, TEXT( "CNPACAudioCoder::AddHeader" ) ));

	// add npac-audio string
	strcpy( (char*)pOut, "npac-audio" );
	pOut += 12;

	// add rate	- 2 bytes
	CMediaType mt = m_pInput->CurrentMediaType();
	*pOut = (BYTE)( ( (WAVEFORMATEX*)mt.pbFormat )->nSamplesPerSec / 256 );
	pOut ++;
	*pOut = (BYTE)LOBYTE( ( (WAVEFORMATEX*)mt.pbFormat )->nSamplesPerSec );
	pOut ++;

	// add channels - 1 byte (1 for mono, 2 for stereo)
	*pOut = (BYTE)( (WAVEFORMATEX*)mt.pbFormat )->nChannels;
	pOut ++;

	// add sample width	- 1 byte (in bits)
	*pOut = (BYTE)( (WAVEFORMATEX*)mt.pbFormat )->wBitsPerSample;
	pOut ++;

	// add format description - 10 bytes (adpcm or gsm)
	strcpy( (char*)pOut, "adpcm" );
	pOut += 10;

	DbgLog(( LOG_TRACE, LOWEST, TEXT( "AddHeader: HEADER CREATED" ) ));
	return;
}


//
// DllRegisterServer
//
HRESULT DllRegisterServer()
{
	return AMovieDllRegisterServer();
}


//
// DllUnregisterServer
//
HRESULT DllUnregisterServer()
{
	return AMovieDllUnregisterServer();
}
