/*********************************************************************
 *
 * File:         application.c
 *
 * Description:  NPAC Conferencing System API functions for applications 
 *               management 
 *
 * Author:       Tomasz Stachowiak, NPAC
 *
 * Date:         10-22-96
 *
 *********************************************************************/
#include <stdio.h>

#include "common.h"

extern int errno;


/* send request to the Conference Engine */
int appEngineRequest(ncsApplication *app, short tag, char *data, int len)
{
  Message *msg;
  int size;
  DWORD user_len;

  size = sizeof(Message) + len - 1;
  msg = (Message *)malloc(size);
  if(msg == NULL)
    {
      fprintf(stderr, "Application: control message allocation error\n");
      return 0;
    }
  msg->local = 1;
  msg->tag = tag;
  memcpy(msg->data, data, len);
  GetUserName(msg->sender.user, &user_len);
  getHostFullName(msg->sender.host, 50);

  if(sendPacket(app->ctrl_socket, &app->daemon_addr, (char *)msg, size) < 0)
    return 0; 

  return 1;
}

void appAddMember(ncsApplication *app, char *data)
{
  MemberInfo *newmember;

  newmember = (MemberInfo *)data;

  PRINTF("Application: member added %s@%s handle %d\n", newmember->part.user, newmember->part.host, newmember->handle);

  app->memberActive[newmember->handle] = 1;
  memcpy(&app->member[newmember->handle], newmember, sizeof(MemberInfo));
  app->member[newmember->handle].addr.sin_port = htons(app->port);

  app->numMembers++;
}

void appDeleteMember(ncsApplication *app, char *data)
{
  int *no;

  no = (int *)data;
  app->memberActive[*no] = 0;
  app->numMembers--;
}

void appProcessMessage(ncsApplication *app, void *data, int len)
{
  Message *msg;
  int size;
  InviteData *info;

  msg = (Message *)data;
  size = len - sizeof(Message) + 1;
  info = (InviteData *)msg->data;
  switch((LocalTag)msg->tag)
    {
    case indLocalSessionAddMember:
      PRINTF("Application: add member\n");
      appAddMember(app, msg->data);
      break;

    case indLocalSessionDeleteMember:
      PRINTF("Application: delete member\n");
      appDeleteMember(app, msg->data);
      break;

    case indLocalSessionLocalMember:
      PRINTF("Application: local member %d\n", *(int *)msg->data);
      app->localHandle = *(int *)msg->data;
      if(app->startingCallback != NULL)
	app->startingCallback();
      break;

    case indLocalAppQuit:
      PRINTF("Application: quit\n");
      if(app->quitCallback != NULL)
	app->quitCallback();
      break;

    default:
      fprintf(stderr, "Application: wrong message tag %d\n", (LocalTag)msg->tag);
    }
}

ncsApplication *ncsAppAttach(int port)
{
  ncsApplication *app;
  char host_name[50];
  struct hostent *host_data;
  int i;
  WSADATA wsaData;	

  if(WSAStartup(257, &wsaData) != 0)
	  return 0;
  app = (ncsApplication *)malloc(sizeof(ncsApplication));
  
  app->port = port;
  
  /* Initialize application data socket */
  if ((app->socket = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
      fprintf(stderr, "Application: can't open socket\n");
      free(app);
      return NULL;
    }
  
  app->addr.sin_family = AF_INET;
  app->addr.sin_addr.s_addr = htonl(INADDR_ANY);
  app->addr.sin_port        = htons(port);
  
  if(bind(app->socket, (SOCKADDR *)&app->addr, sizeof(struct sockaddr_in)) < 0)
    {
      fprintf(stderr, "Application: can't bind socket\n");
      free(app);
      return NULL;
    }

  /* Initialize application control socket */
  if ((app->ctrl_socket = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
      fprintf(stderr, "Application: can't open control socket\n");
      free(app);
      return NULL;
    }

  app->ctrl_addr.sin_family = AF_INET;
  app->ctrl_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  app->ctrl_addr.sin_port        = htons(port+1);
  
  if(bind(app->ctrl_socket, (SOCKADDR *)&app->ctrl_addr, sizeof(struct sockaddr_in)) < 0)
    {
      fprintf(stderr, "Application: can't bind control socket\n");
      free(app);
      return NULL;
    }

  /* Set engine address */
  app->daemon_addr.sin_family = AF_INET;
  getHostFullName(host_name, 50);
  host_data = gethostbyname(host_name);
  app->daemon_addr.sin_addr.s_addr = *(long *)host_data->h_addr;
  app->daemon_addr.sin_port        = htons(DAEMON_PORT);

  /* Send request to engine */
  appEngineRequest(app, reqLocalApplicationAttach, 
		   (char *)&app->ctrl_addr, sizeof(struct sockaddr_in));


  for(i=0; i<MAX_MEMBERS; i++)
	  app->memberActive[i] = 0;

  app->numMembers = 0;
  app->startingCallback = NULL;
  app->quitCallback = NULL;

  return app;
}

void ncsAppDettach(ncsApplication *app)
{
  if(app != NULL)
    {
      /* Send request to engine */
      appEngineRequest(app, reqLocalApplicationDettach, 
		       (char *)&app->port, sizeof(int));
      closesocket(app->socket);
      closesocket(app->ctrl_socket);
      free(app);
    }
}

int ncsAppSendData(ncsApplication *app, int mbr, DWORD type, void *data, int len)
{
  ncsData *ncsdata;
  int size;
  DWORD i;
  int sent = 0;
	char text[100];

  if(len == 0)
    {
      size = sizeof(ncsData);
      ncsdata = (ncsData *)malloc(size);
      ncsdata->data[0] = 0;
    }
  else
    {
      size = len + sizeof(ncsData) - 1;
      ncsdata = (ncsData *)malloc(size);
      memcpy(ncsdata->data, data, len); 
    }

  ncsdata->type = htonl(type);
  ncsdata->handle = htonl(app->localHandle);

  if(mbr < 0)  /* send to all members */
  {
	for(i=0; i<MAX_MEMBERS; i++)
		if(app->memberActive[i])
		{
			if(i != app->localHandle)
			{
				PRINTF("Application: data sent to user %d, size %d\n", i, size);
				app->member[i].addr.sin_port = htons(app->port);
				sendPacket(app->socket, &app->member[i].addr, (char *)ncsdata, size);
			}
		}
  }
  else   /* send to one specific member */
    {
      if(app->memberActive[mbr])
	if(mbr != app->localHandle)
	  {
	    app->member[mbr].addr.sin_port = htons(app->port);
	    sendPacket(app->socket, &app->member[mbr].addr, (char *)ncsdata, size);
	  }

      /* if message to initiator find initiator */
      if((mbr == 0) && (app->memberActive[mbr] == 0))
	{
	  i = 0;
	  while((i < MAX_MEMBERS) && (!sent))
	    {
	      if(i != app->localHandle)
		{
		  PRINTF("Application: data sent to user %d, size %d\n",i, size);
		  app->member[i].addr.sin_port = htons(app->port);
		  sendPacket(app->socket, &app->member[i].addr, (char *)ncsdata, size);
		  sent = 1;
		}
	      i++;
	    }
	}
    }      

  free(ncsdata);

  return 0;
}

/*************************************************************************
 * 
 * Proccesses the NCS message incoming from the application control  
 * socket (from the Conference Engine) 
 *
 *************************************************************************/ 
int ncsAppProcessIncoming(ncsApplication *app, fd_set *set)
{
  int msg_len, n, remote_addr_len;
  struct sockaddr_in remote_addr;
  char *data;

  remote_addr_len = sizeof(struct sockaddr_in);
  if(FD_ISSET(app->ctrl_socket, set))
    {
      data = (char *)malloc(MAX_UDP_SIZE);
      n = recvfrom(app->ctrl_socket, data, MAX_UDP_SIZE, 
		   0, (SOCKADDR *)&remote_addr, &remote_addr_len); 
      if (n < 0)
	{
	  fprintf(stderr, "Application: recvfrom error\n");
	  return 0;
	}
      else
	{
	  PRINTF("Application: received message size %d\n", n);
	  appProcessMessage(app, data, n); 
	}

      free(data);
    }
  else
    return 0;

  return 1;
}

/*************************************************************************
 * 
 * Receive the application data(if any) from the application data  
 * socket (from remote applications) 
 *
 *************************************************************************/ 
int ncsAppGetData(ncsApplication *app, fd_set *set, char **data)
{
  int msg_len, n, remote_addr_len;
  struct sockaddr_in remote_addr;

  remote_addr_len = sizeof(struct sockaddr_in);
  if(FD_ISSET(app->socket, set))
    {
      PRINTF("Application: there are data\n");

      *data = (char *)malloc(MAX_UDP_SIZE);
      n = recvfrom(app->socket, (char *)*data, MAX_UDP_SIZE, 
		   0, (SOCKADDR *)&remote_addr, &remote_addr_len); 
      if (n < 0)
	{
	  fprintf(stderr, "Application: recvfrom error\n", errno);
	  return 0;
	}
    }
  else
    return 0;

  return n;
}  


/***********************************************************************
 * 
 * Get the last file descriptor used by NCS and set the file descriptors
 * set 'set' for all descriptors used by NCS
 *
 ***********************************************************************/
int ncsAppGetFD(ncsApplication *app, fd_set *set)
{
  FD_ZERO(set);
  FD_SET(app->socket, set);
  FD_SET(app->ctrl_socket, set);
  return app->ctrl_socket;
}

DWORD ncsAppGetLocalHandle(ncsApplication *app)
{
  return app->localHandle;
}

int ncsAppGetNumMembers(ncsApplication *app)
{
  return app->numMembers;
}

void ncsAppAddCallback(ncsApplication *app, short tag, void (* function))
{
  switch(tag)
    {
    case tagAppStarting:
      app->startingCallback = (void (*)(void))function;
      break;

    case tagAppQuit:
      app->quitCallback = (void (*)(void))function;
      break;

    default:
      fprintf(stderr, "Unknown application event tag\n");
    }
}
	













