/*********************************************************************
 *
 * File:         sessions.cpp
 *
 * Description:  NPAC Conferencing System API functions for session 
 *               management 
 *
 * Author:       Tomasz Stachowiak, NPAC
 *
 * Date:         10-14-96
 *
 *********************************************************************/
#include <stdio.h>

#include "common.h"
#include "session.h"

MemberInfo member[MAX_MEMBERS];
int memberActive[MAX_MEMBERS];
int localHandle;

void (* inviteCallback)(void *);
void (* deleteMemberCallback)(void *);
void (* addMemberCallback)(void *);
void (* startAppCallback)(void *);
void (* serverUsersCallback)(void *);

ManagerData managerData;


void InitMembers()
{
  int i;

  for(i=0; i<MAX_MEMBERS; i++)
    memberActive[i] = 0;
}

void AddMember(char *data)
{
  MemberInfo *newmember;

  newmember = (MemberInfo *)data;
  PRINTF("Member added locally %s@%s handle %d\n", newmember->part.user, newmember->part.host, newmember->handle);
  memberActive[newmember->handle] = 1;
  memcpy(&member[newmember->handle], newmember, sizeof(MemberInfo));

  addMemberCallback(&newmember->part);
}

void DeleteMember(char *data)
{
  int *no;

  no = (int *)data;
  memberActive[*no] = 0;

  deleteMemberCallback(&member[*no].part);
}

void printMembers(InviteData *data, int len)
{
  int no, i;

  no = data->numMembers;
  printf("Conference members:\n");
  for(i=0; i<no; i++)
    printf("%s@%s\n", data->parts[i].user, data->parts[i].host);
}

int ncsSessionInitialize()
{
  char host_name[50];
  struct hostent *host_data;
  WSADATA wsaData;	

  if(WSAStartup(257, &wsaData) != 0)
	  return 0;
  
  if ((managerData.socket = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    {
      fprintf(stderr, "Manager: can't open manage socket\n");
      return 0;
    }

  managerData.addr.sin_family = AF_INET;
  managerData.addr.sin_addr.s_addr = htonl(INADDR_ANY);
  managerData.addr.sin_port        = htons(MANAGER_PORT);

  if(bind(managerData.socket, (SOCKADDR *)&managerData.addr, 
	  sizeof(managerData.addr)) < 0)
    {
      fprintf(stderr, "Manager: can't bind daemon socket\n");
      return 0;
    }

  managerData.daemon_addr.sin_family = AF_INET;
  getHostFullName(host_name, 50);
  if((host_data = gethostbyname(host_name)) == NULL)
	  MessageBox(NULL, "Invalid host", "Error", MB_OK);
  managerData.daemon_addr.sin_addr.s_addr = *(long *)host_data->h_addr;
  managerData.daemon_addr.sin_port        = htons(DAEMON_PORT);
  
  ncsEngineRequest(reqLocalSessionInit, NULL, 0);

  InitMembers();

  return 1;
}

void ncsSessionCreate()
{
  ncsEngineRequest(reqLocalSessionCreate, NULL, 0);
  localHandle = 0;     /* what if session already created?!!! */
}



/* Registers working procedure running in the Xt background, 
   processing incoming control messages */
void ncsSessionRegisterWithXt(/*XtApp app*/)
{
}

/* Returns file descriptor to session control socket */
int ncsSessionGetFD(fd_set *set)
{
  FD_ZERO(set);
  FD_SET(managerData.socket, set);
  return managerData.socket;
}

/* Processes control messages coming to the control socket */
void ncsProcessIncoming(fd_set *set)
{
  char *msg;
  int n, msg_len;
  struct sockaddr_in addr;
  int addr_size;
  
  addr_size = sizeof(addr);
  if(FD_ISSET(managerData.socket, set))
    {
      msg = (char *)malloc(MAX_UDP_SIZE);
      n = recvfrom(managerData.socket, msg, MAX_UDP_SIZE, 0, (SOCKADDR *)&addr, &addr_size); 
      if (n < 0)
	fprintf(stderr, "Manager: recvfrom error\n");
      
      ncsEngineMessageHandle(msg, n);
      
      free(msg);
    }
}


void ncsEngineMessageHandle(void *message, int len)
{
  Message *msg;
  InviteData *info;
  int size;

  msg = (Message *)message;
  size = len - sizeof(Message) + 1;
  info = (InviteData *)msg->data;

  if(((ServerTag)ntohs(msg->tag)) == indServerUsers)
	{
		if(serverUsersCallback != NULL)
			serverUsersCallback(msg->data);
		return;
	}

  switch((LocalTag)msg->tag)
    {
    case indLocalSessionInvited:
      inviteCallback((Participant *)msg->data);
      break;

    case indLocalSessionAccept:
      PRINTF("Manager: Invitation accepted\n");
      break;

    case indLocalSessionReject:
      PRINTF("Manager: Invitation rejected\n");
      break;
      
    case indLocalSessionMembers:
      printMembers(info, size);
      break;

    case indLocalSessionAddMember:
      PRINTF("Manager: add member\n");
      AddMember(msg->data);
      break;

    case indLocalSessionDeleteMember:
      DeleteMember(msg->data);
      break;

    case indLocalSessionLocalMember:
      localHandle = *(int *)msg->data;
      break;

    case indLocalSessionStartApp:
      PRINTF("Manager: start application\n");
      if(startAppCallback != NULL)
		startAppCallback(msg->data);
      break;

    default:
      fprintf(stderr, "Manager: wrong message tag %d\n", (LocalTag)msg->tag);
    }
}

void ncsSessionAddMember(Participant *part)
{
  ncsEngineRequest(reqLocalSessionAddMember, (char *)part, sizeof(Participant));
}

void ncsSessionAcceptInvitation(Participant *part)
{
  PRINTF("Manager: join session %s@%s \n", part->user, part->host);
  ncsEngineRequest(reqLocalSessionJoin, (char *)part, sizeof(Participant));
}

void ncsSessionRejectInvitation(Participant *part)
{
  ncsEngineRequest(reqLocalSessionReject, (char *)part, sizeof(Participant));
}

void ncsSessionAddCallback(short tag, void (*function))
{
  switch(tag)
    {
    case indLocalSessionInvited:
      inviteCallback = (void (*)(void *))function;
      break;

    case indLocalSessionAddMember:
      addMemberCallback = (void (*)(void *))function;
      break;

    case indLocalSessionDeleteMember:
      deleteMemberCallback = (void (*)(void *))function;
      break;

    case indLocalSessionStartApp:
      startAppCallback = (void (*)(void *))function;
      break;

    case indServerUsers:
      serverUsersCallback = (void (*)(void *))function;
      break;

    default:
      fprintf(stderr, "Manager: Unknown control callback type\n");
    }
}

void ncsSessionQuit()
{
  ncsEngineRequest(reqLocalSessionQuit, NULL, 0);
  closesocket(managerData.socket);
}

void ncsSessionQueryMembers()
{
  ncsEngineRequest(reqLocalSessionGetMembers, NULL, 0);
}


void ncsSessionSendToApp(int port, short type, char *msg, int len)
{
  AppData *data;
  int size;

  if(len == 0)
    {
      size = sizeof(AppData);
      data = (AppData *)malloc(size);
      data->data[0] = 0;
    }
  else
    {
      size = sizeof(AppData) + len - 1;
      data = (AppData *)malloc(size);
      memcpy(data->data, data, len);
    }
    

  data->port = port + 1;
  data->tag = type;
  memcpy(data->data, msg, len);
  ncsEngineRequest(reqLocalSessionSendToApp, (char *)data, size);
  free(data);
}

void ncsSessionStartApp(int port)
{
  ncsEngineRequest(reqLocalSessionStartApp, (char *)&port, sizeof(int));
}








